#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to fill a layer up to a given pixel density using the current brush
# (c) Ofnuts 2010
#
#   History:
#
#   v0.0: 2011-11-19: First published version
#   v0.1: 2011-11-27: Fix undo behavior
#   v0.2: 2012-12-05: Remove print instructions
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import math,random,os,sys
from gimpfu import *

class Filler(object):
	def __init__(self,image,refLayer,density):
		# Get brush data
		brush=pdb.gimp_context_get_brush()
		brushW,brushH,mask_bpp,num_mask_bytes,mask_bytes,color_bpp,num_color_bytes,color_bytes=pdb.gimp_brush_get_pixels(brush)
		# Warning; mask_bpp assumed to be 1!
		self.brushAllPixels=num_mask_bytes
		#print mask_bytes
		self.brushActualPixels=sum(mask_bytes)/255
		#print 'Brush: %s: %d x %d = %d, actual %3.2f (%3.2f%%)' % (brush,brushW,brushH, self.brushAllPixels,self.brushActualPixels,100*self.brushActualPixels/self.brushAllPixels)
		self.brushPixels=self.brushActualPixels
		
		# Create layer where we paint
		self.image=image
		self.layer=gimp.Layer(self.image, '%s filled @%d%% with %s' % (refLayer.name,density,brush), refLayer.width, refLayer.height,RGBA_IMAGE, 100, NORMAL_MODE)
		self.image.add_layer(self.layer,0)
		x,y=refLayer.offsets
		self.layer.set_offsets(x,y)
		
		# compute bounds for brush strokes
		selection,minX,minY,maxX,maxY=pdb.gimp_drawable_mask_bounds(self.layer);
		self.minX=minX-brushW/2
		self.maxX=(maxX+brushW/2)-1
		self.minY=minY-brushH/2
		self.maxY=(maxY+brushH/2)-1
		#print 'Bounds: %s,%s -> %s,%s' % (self.minX,self.minY,self.maxX,self.maxY)
		
		# compute number of pixels to cover
		mean, std_dev, median, selectionPixels, count, percentile = pdb.gimp_histogram(self.layer, HISTOGRAM_ALPHA,0,255)
		self.targetPixels=selectionPixels*density/100
		self.coveredPixels=0
		#print 'Target pixels: %s * %s%% = %s' % (selectionPixels,density,self.targetPixels)
		
		self.iterations=0
	
	def randomPosition(self):
		x=random.randrange(self.minX,self.maxX)
		y=random.randrange(self.minY,self.maxY)
		return [x,y]
	
	def paintSpots(self,strokes):
		for i in range(strokes):
			pdb.gimp_paintbrush_default(self.layer,4,self.randomPosition()*2)
			
	
	def fillSome(self):
		#print 'Executed %d iterations' % self.iterations
		# Avoid infinite loops for high densities
		if self.iterations >= 100:
			#print 'Iteration limit reached'
			return False
		self.iterations=self.iterations+1
	
		remainingPixels=self.targetPixels - self.coveredPixels
		# our true stopping condition: next round would be above limit
		if (remainingPixels < self.brushPixels/2):
			#print 'Density reached'
			return False

		# compute number of strokes, then be a bit on the low side to avoid
		# overshoot
		strokes=int(.75*(remainingPixels/self.brushPixels))
		strokes=max(1,strokes)
		
		#print 'Run: %2.2f -> %2.2f (%2.2f), %d@%2.2f' % (self.coveredPixels,self.targetPixels,remainingPixels,strokes,self.brushPixels)
		self.paintSpots(strokes)
		previousPixels=self.coveredPixels
		mean, std_dev, median, pixels, self.coveredPixels, percentile = pdb.gimp_histogram(self.layer, HISTOGRAM_ALPHA,1,255)
		newPixels=self.coveredPixels-previousPixels
		brushPixels=newPixels / strokes
		#print 'Covered pixels: %2.2f / %2.2f: %2.2f new, %2.2f pixels/stroke' % (self.coveredPixels,self.targetPixels,newPixels,brushPixels)
		# reevaluate coverage for next run
		if (brushPixels > self.brushActualPixels/10):
			self.brushPixels=brushPixels
		return True
	   
def densityBrushFill(image,drawable,density):
	# save stuff
	image.undo_group_start()
	gimp.context_push()
	try:
		f=Filler(image,drawable,density)
		image.undo_freeze()
		while f.fillSome():
			pass
		image.undo_thaw()
		
        except Exception as e:
		#print '%s: %s' % (type(e),e)
		pdb.gimp_message('%s: %s' % (type(e),e))
		
	# restore stuff
	gimp.context_pop()
	image.undo_group_end()
	pdb.gimp_displays_flush()
	return;


### Registrations
whoiam='\n'+os.path.abspath(sys.argv[0])
    
register(
    "density-brush-fill",
    N_("Fill with brush strokes to density"+whoiam),
    "Fills with strokes of the current brush to required density",
    "Ofnuts",
    "Ofnuts",
    "2011",
    N_("Fill with brush strokes to density..."),
    "RGB*, GRAY*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
	(PF_SPINNER, "density", "Density", 50, (0, 100, 1)),
    ],
    [],
    densityBrushFill,
    menu="<Image>/Filters/Render",
    domain=("gimp20-python", gimp.locale_directory)
)

main()
